/*
 * Cluster.java
 *
 * Copyright (c) 2011 Daniel Wayne Crabtree - All Rights Reserved
 * 
 * If you use this, please cite my PhD Thesis.
 
   @PHDTHESIS{IWS11,
       AUTHOR =  {Daniel Crabtree},
       TITLE =   {Techniques for Improving Web Search by Understanding Queries},
       SCHOOL =  {School of Engineering and Computer Science, Victoria University of Wellington},
       ADDRESS = {Wellington, New Zealand},
       MONTH =   {May},    
       YEAR =    {2011}
   }
  
 */

package qc4;

import java.util.*;

public class Cluster implements Cloneable, Iterable<Page>
{   
    private BitSet pages;
    private int pagesize;
    private ResultSet resultSet;
        
    public Cluster(ResultSet resultSet)
    {
        this.pages = new BitSet(resultSet.size());
        this.resultSet = resultSet;
        this.pagesize = 0;
    }
        
    public Object clone()
    {
        Cluster object = null;
        
        try
        {
            object = (Cluster)super.clone();
            object.pages = (BitSet)this.pages.clone();
            object.pagesize = this.pagesize;
            object.resultSet = this.resultSet;
        }
        catch(CloneNotSupportedException ex)
        {
            System.err.println("Clone not supported in Cluster");
        }
        
        return object;
    }
    
    public boolean equals(Cluster cluster)
    {
        return this.intersectionSize(cluster) == this.size();
    }
    
    public Iterator<Page> iterator()
    {
        return new ClusterIterator(this);
    }
    
    public final boolean contains(Page page)
    {
        return this.pages.get(page.getLocation());
    }
           
    public final int size()
    {
        if(this.pagesize == -1)
            this.pagesize = this.pages.cardinality();
                
        return this.pagesize;
    }    
    
    public final void addPage(int location)
    {
        this.addPage(this.resultSet.getPage(location));
    }
    
    public final void addPageRange(int start, int end)
    {
        for(Page page: this.resultSet.getPageRange(start, end))
        {
            this.addPage(page);
        }
    }
    
    private void addPage(Page page)
    {
        int location = page.getLocation();
        if(!this.pages.get(location))
        {
            this.pagesize++;
            this.pages.set(location);
        }
    }
                
    final Page getPage(int index)
    {
        if(this.pages.get(index))
            return this.resultSet.getPage(index);

        return null;
    }
    
    final int nextIndex(int index)
    {
        return this.pages.nextSetBit(index);
    }
    
    final <T extends Cluster> boolean isPredominantly(Cluster cluster, Iterable<T> clustering)
    {
        if(cluster == null)
            return false;
        
        int intersectionSize = this.intersectionSize(cluster);
        for(Cluster cluster2: clustering)
        {
            if(!cluster.equals(cluster2) && intersectionSize < this.intersectionSize(cluster2))
                return false;
        }
        
        return true;
    }
    
    final int intersectionSize(Cluster cluster)
    {
        BitSet intersection = (BitSet)this.pages.clone();
        intersection.and(cluster.pages);
        return intersection.cardinality();
    }

    final Cluster intersection(Cluster cluster)
    {
        Cluster output = (Cluster)this.clone();
        output.pagesize = -1;
        output.pages.and(cluster.pages);
        return output;
    }
    
    final Cluster difference(Cluster cluster)
    {
        Cluster output = (Cluster)this.clone();
        output.pagesize = -1;
        output.pages.andNot(cluster.pages);
        return output;
    }
    
    final boolean isSubCluster(Cluster cluster)
    {
        BitSet subcluster = (BitSet)this.pages.clone();
        subcluster.and(cluster.pages);
        return subcluster.equals(this.pages);
    }
}