/*
 * QC4.java
 *
 * Copyright (c) 2011 Daniel Wayne Crabtree - All Rights Reserved
 * 
 * If you use this, please cite my PhD Thesis.
 
   @PHDTHESIS{IWS11,
       AUTHOR =  {Daniel Crabtree},
       TITLE =   {Techniques for Improving Web Search by Understanding Queries},
       SCHOOL =  {School of Engineering and Computer Science, Victoria University of Wellington},
       ADDRESS = {Wellington, New Zealand},
       MONTH =   {May},    
       YEAR =    {2011}
   }
  
 */

package qc4;

import java.util.*;

public class QC4 {
    
    private IdealClustering idealClustering;
    private int digitsAfterDecimalPoint;
    private Map<Cluster, Integer> clusterLevelCache; 
    
    public QC4(IdealClustering idealClustering, int digitsAfterDecimalPoint)
    {
        this.idealClustering = idealClustering;
        this.digitsAfterDecimalPoint = digitsAfterDecimalPoint;
        this.clearCache();
    }
    
    private void clearCache()
    {
        this.clusterLevelCache = new HashMap<Cluster, Integer>();
    }
    
    public final float AQ(Clustering<Cluster> clustering)
    {
        this.clearCache();
        
        Fraction result = new Fraction();
        for(Cluster cluster: clustering.clustersExcludingMostlyOutliers(this.idealClustering))
        {
            result.top += this.QU(cluster);
            result.bottom++;
        }
                
        return result.rounded();
    }
    public final float WQ(Clustering<Cluster> clustering)
    {
        this.clearCache();
        
        Fraction result = new Fraction();
        for(Cluster cluster: clustering.clustersExcludingMostlyOutliers(this.idealClustering))
        {
            result.top += this.QU(cluster) * cluster.size();
            result.bottom += cluster.size();
        }
        
        return result.rounded();
    }   
    public final float AC(Clustering<Cluster> clustering)
    {
        this.clearCache();
        
        Fraction result = new Fraction();
        for(IdealCluster topic: this.idealClustering.clustersExcludingOutlierOnLevel(1))
        {
            result.top += this.CV(clustering, topic);
            result.bottom++;
        }
        
        return result.rounded();
    }
    public final float WC(Clustering<Cluster> clustering)
    {
        this.clearCache();
        
        Fraction result = new Fraction();
        for(IdealCluster topic: this.idealClustering.clustersExcludingOutlierOnLevel(1))
        {
            result.top += this.CV(clustering, topic) * topic.size();
            result.bottom += topic.size();
        }

        return result.rounded();
    }
    
    private float QU(Cluster cluster)
    {       
        return (1 - this.modifiedEntropy(cluster)) * this.min(1, this.sRandom(cluster), this.sSegment(cluster));
    }
    private float modifiedEntropy(Cluster cluster)
    {
        float modifiedEntropy = Float.MAX_VALUE;
        for(IdealCluster topic: this.idealClustering.clustersOnLevel(this.clusterLevel(cluster)))
        {
            float value = this.entropyWithModifiedPrecision(cluster, topic);
            if(value < modifiedEntropy)
                modifiedEntropy = value;
        }
        return modifiedEntropy;
    }
    private float entropyWithModifiedPrecision(Cluster cluster, IdealCluster bestTopic)
    {
        List<IdealCluster> clustersOnLevel = this.idealClustering.clustersOnLevel(this.clusterLevel(cluster));       
        int numberAtLevel = clustersOnLevel.size();
        float entropy = 0;
        for(IdealCluster topic: clustersOnLevel)
        {
            float precision = this.modifiedPrecision(cluster, topic, bestTopic);
            if(precision > 0 && numberAtLevel > 1)
                entropy += precision * Math.log(precision) / Math.log(numberAtLevel);
        }
        
        return -1 * entropy;
    }
    
    private float modifiedPrecision(Cluster cluster, IdealCluster topic, IdealCluster bestTopic)
    {
        float output = 0;
        if(topic.equals(bestTopic))
            output = this.precision(cluster, topic);
        else
            output = this.normalizedPrecision(cluster, topic, bestTopic);

        return output;
    }
    private float normalizedPrecision(Cluster cluster, IdealCluster topic, IdealCluster bestTopic) 
    {
        Fraction fraction = new Fraction();
        fraction.top = this.normalizedPrecisionNumerator(cluster, topic, bestTopic);
        fraction.bottom = this.normalizedPrecisionDenominator(cluster, bestTopic);
        return fraction.exact();
    }
    private float normalizedPrecisionNumerator(Cluster cluster, IdealCluster topic, IdealCluster bestTopic) 
    {
        Cluster inClusterAndTopic = topic.intersection(cluster);
        Cluster inClusterAndBestTopic = bestTopic.intersection(cluster);
        Cluster inClusterAndTopicButNotBestTopic = inClusterAndTopic.difference(inClusterAndBestTopic);
        return (cluster.size() - inClusterAndBestTopic.size()) * inClusterAndTopicButNotBestTopic.size();
    }    
    private float normalizedPrecisionDenominator(Cluster cluster, IdealCluster bestTopic) 
    {
        float bottom = 0;
        Cluster inClusterAndBestTopic = bestTopic.intersection(cluster);
        for(IdealCluster topic: this.idealClustering.clustersOnLevel(this.clusterLevel(cluster)))
        {
            if(!topic.equals(bestTopic))
            {
                Cluster inClusterAndTopic = topic.intersection(cluster);
                Cluster inClusterAndTopicButNotBestTopic = inClusterAndTopic.difference(inClusterAndBestTopic);
                
                bottom += inClusterAndTopicButNotBestTopic.size();
            }
        }
        bottom *= cluster.size();
        return bottom;
    }
    
    private float sRandom(Cluster cluster)
    {     
        List<Cluster> regionsAtClusterLevel = this.idealClustering.getRegions(this.clusterLevel(cluster));
        if(regionsAtClusterLevel.size() > 1)
            return (1F - uniformity(cluster, regionsAtClusterLevel)) / 0.05F;
        
        return 1;
    }
    private float uniformity(Cluster cluster, List<Cluster> regions)
    {
        float uniformity = 0;

        List<Float> recallValues = new ArrayList<Float>();
        float recallSum = 0;

        for(Cluster region: regions)
        {
            float numberInIntersection = cluster.intersectionSize(region);
            if(numberInIntersection > 0)
            {
                float recall = numberInIntersection / region.size();
                recallValues.add(recall);
                recallSum += recall;
            }
        }

        for(Float recall: recallValues)
        {
            float normalizedRecall = recall / recallSum;
            uniformity += normalizedRecall * Math.log(normalizedRecall) / Math.log(regions.size());
        }
        
        return -1 * uniformity;
    }
    
    private float sSegment(Cluster cluster)
    {
        float isize = -1;
        int clusterlevel = this.clusterLevel(cluster);        
        for(IdealCluster bestTopic: this.idealClustering.clustersOnLevel(clusterlevel))
        {
            float value = 0;
            for(IdealCluster topic: this.idealClustering.clustersOnLevel(clusterlevel))
            {
                value += this.modifiedPrecision(cluster, topic, bestTopic) * this.s(cluster, topic);
            }

            if(isize == -1 || value > isize)
                isize = value;
        }
        
        if(isize == -1)
            return 1;
        
        return isize;
    }
    private float s(Cluster cluster, IdealCluster topic)
    {       
        float recall = this.recall(cluster, topic);
        return (float)Math.pow(2, (recall - 1) / (recall * Math.log(topic.size()) / Math.log(2)));
    }
   
    private float CV(Clustering<Cluster> clustering, IdealCluster topic)
    {
        float sum = 0;
        for(Page page: topic)
            sum += this.pageCoverage(clustering, page, topic, 1);

        return sum / (float)topic.size();
    }
    private float pageCoverage(Clustering<Cluster> clustering, Page page, IdealCluster topic, int level)
    {
        Fraction result = new Fraction();
        for(IdealCluster idealCluster: this.idealClustering.clustersOnLevel(level))
        {           
            if(idealCluster.contains(page) && topic.hasDescendant(idealCluster))
            {
                result.top += Math.max(this.individualPageCoverage(clustering, page, idealCluster, level),
                                       this.pageCoverage(clustering, page, idealCluster, level+1));
                result.bottom++;
            }
        }
        
        return result.exact();
    }
    private float individualPageCoverage(Clustering<Cluster> clustering, Page page, IdealCluster topic, int level)
    {
        float max = 0;
        for(Cluster cluster: clustering.clustersExcludingMostlyOutliers(this.idealClustering))
        {
            if(cluster.contains(page) && this.clusterLevel(cluster) == level)
            {
                float precision = this.precision(cluster, topic) *
                                  this.min(1, this.sRandom(cluster), this.sTranche(cluster, clustering));
                if(precision > max)
                    max = precision;
            }
        }
        return max;
    }
    
    private float sTranche(Cluster cluster, Clustering<Cluster> clustering)
    {
        int clusterRank = this.trancheRank(this.tranche(cluster), clustering);
        if(clusterRank <= Math.max(15, this.idealClustering.size()))
            return 1F;
        
        return (float)cluster.size() / (float)clusterRank;
    }
    private int tranche(Cluster cluster)
    {
        return (int)Math.ceil(Math.log(cluster.size()) / Math.log(4));
    }
    private int trancheRank(int tranche, Clustering<Cluster> clustering)
    {
        int rank = 0;
        for(int i = tranche; i <= this.trancheMax(clustering); i++)
        {
            rank += this.trancheSize(i, clustering);
        }
        return rank;
    }    
    private int trancheMax(Clustering<Cluster> clustering)
    {
        int max = 0;
        for(Cluster cluster: clustering)
        {
            int tranche = this.tranche(cluster);
            if(tranche > max) 
                max = tranche;
        }
        return max;
    }
    private int trancheSize(int tranche, Clustering<Cluster> clustering)
    {
        int count = 0;
        for(Cluster cluster: clustering)
        {
            if(this.tranche(cluster) == tranche) 
                count++;
        }
        return count;
    }
    
    private int clusterLevel(Cluster cluster)
    {
        if(!this.clusterLevelCache.containsKey(cluster))
        { 
            int maxLevel = 1;
            float maxF = -1;

            for(IdealCluster topic: this.idealClustering.clustersExcludingOutlier())
            {
                float f = this.fMeasure(cluster, topic);
                if(maxF == -1 || f > maxF || (f == maxF && topic.getLevel() > maxLevel))
                {
                    maxLevel = topic.getLevel();
                    maxF = f;
                }
            }
            
            this.clusterLevelCache.put(cluster, maxLevel);
        }
        
        return this.clusterLevelCache.get(cluster);
    }
    private float fMeasure(Cluster cluster, IdealCluster topic)
    {
        float p = this.precision(cluster, topic);
        float r = this.recall(cluster, topic);
        
        if(p + r > 0)
            return (2 * p * r) / (p + r);
        
        return 0; 
    }
    private float precision(Cluster cluster, IdealCluster topic)
    {
        return (float)topic.intersectionSize(cluster) / (float)cluster.size();
    }
    private float recall(Cluster cluster, IdealCluster topic)
    {
        return (float)topic.intersectionSize(cluster) / (float)topic.size();
    }
    
    private float min(float v1, float v2, float v3)
    {
        return Math.min(Math.min(v1, v2), v3);
    }
    
    private class Fraction
    {
        float top = 0;
        float bottom = 0;
        
        float rounded()
        {
            return this.round(this.exact());
        }
        
        float exact()
        {
            float value = 0;
            if(this.bottom > 0)
                value = this.top / this.bottom;
            
            return value;
        }
        
        private float round(float number)
        {
            float significantDigitsBeforeDecimalPoint = number * (float)Math.pow(10, digitsAfterDecimalPoint);

            float significantDigits = (float)Math.floor(significantDigitsBeforeDecimalPoint);
            float insignificantDigits = significantDigitsBeforeDecimalPoint - significantDigits;

            if(insignificantDigits >= 0.5)
                significantDigits++;

            return significantDigits / (float)Math.pow(10, digitsAfterDecimalPoint);
        }
    }
}